import os
import re
import shutil
import subprocess
import shlex
import logging
import random
import string
from string import Template
import sys

import riscof.utils as utils
import riscof.constants as constants
from riscof.pluginTemplate import pluginTemplate

logger = logging.getLogger()

class urv(pluginTemplate):
    __model__ = "urv"
    __version__ = "1.0.0"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        config = kwargs.get('config')
        self.config_dir = kwargs.get('config_dir')

        # If the config node for this DUT is missing or empty. Raise an error. At minimum we need
        # the paths to the ispec and pspec files
        if config is None:
            print("Please enter input file paths in configuration.")
            raise SystemExit(1)

        # In case of an RTL based DUT, this would be point to the final binary executable of your
        # test-bench produced by a simulator (like verilator, vcs, incisive, etc). In case of an iss or
        # emulator, this variable could point to where the iss binary is located. If 'PATH variable
        # is missing in the config.ini we can hardcode the alternate here.


        # Number of parallel jobs that can be spawned off by RISCOF
        # for various actions performed in later functions, specifically to run the tests in
        # parallel on the DUT executable. Can also be used in the build function if required.
        self.num_jobs = str(config['jobs'] if 'jobs' in config else 1)

        # Path to the directory where this python file is located. Collect it from the config.ini
        self.pluginpath=os.path.abspath(config['pluginpath'])

        # Collect the paths to the  riscv-config absed ISA and platform yaml files. One can choose
        # to hardcode these here itself instead of picking it from the config.ini file.
        self.isa_spec = os.path.abspath(config['ispec'])
        self.platform_spec = os.path.abspath(config['pspec'])

        #We capture if the user would like the run the tests on the target or
        #not. If you are interested in just compiling the tests and not running
        #them on the target, then following variable should be set to False
        if 'target_run' in config and config['target_run']=='0':
            self.target_run = False
        else:
            self.target_run = True

    def initialise(self, suite, work_dir, archtest_env):

       # capture the working directory. Any artifacts that the DUT creates should be placed in this
       # directory. Other artifacts from the framework and the Reference plugin will also be placed
       # here itself.
       self.work_dir = work_dir
       logger.info(f"work_dir={work_dir}")

       # capture the architectural test-suite directory.
       self.suite_dir = suite

       # Note the march is not hardwired here, because it will change for each
       # test. Similarly the output elf name and compile macros will be assigned later in the
       # runTests function
       self.compile_cmd = 'riscv{1}-elf-gcc -march={0} \
         -static -mcmodel=medany -fvisibility=hidden -nostdlib -nostartfiles -g\
         -T '+self.pluginpath+'/env/link.ld\
         -I '+self.pluginpath+'/env/\
         -I ' + archtest_env + ' {2} -o {3} {4}'

       # Build the Verilator model of the DUT
       build_urv = 'fusesoc run --build-root riscof_dut --target verilator_tb --build ohwr:urv:riscof'
       utils.shellCommand(build_urv).run()

       self.dut_exe = os.path.join(
            self.config_dir,
            'riscof_dut/ohwr_urv_riscof_0/verilator_tb-verilator/Vriscof_tb'
        )

    def build(self, isa_yaml, platform_yaml):

      # load the isa yaml as a dictionary in python.
      ispec = utils.load_yaml(isa_yaml)['hart0']

      # capture the XLEN value by picking the max value in 'supported_xlen' field of isa yaml. This
      # will be useful in setting integer value in the compiler string (if not already hardcoded);
      self.xlen = ('64' if 64 in ispec['supported_xlen'] else '32')

      # for urv start building the '--isa' argument. the self.isa is dutnmae specific and may not be
      # useful for all DUTs
      self.isa = 'rv' + self.xlen
      if "I" in ispec["ISA"]:
          self.isa += 'i'
      if "M" in ispec["ISA"]:
          self.isa += 'm'
      if "F" in ispec["ISA"]:
          self.isa += 'f'
      if "D" in ispec["ISA"]:
          self.isa += 'd'
      if "C" in ispec["ISA"]:
          self.isa += 'c'

      self.compile_cmd = self.compile_cmd+' -mabi='+('lp64 ' if 64 in ispec['supported_xlen'] else 'ilp32 ')

    def runTests(self, testList):

        # We will iterate over each entry in the testList. Each entry node will be referred to by the
        # variable testname.
        for testname in testList:

            logger.debug('Running Test: {0} on DUT'.format(testname))
            # for each testname we get all its fields (as described by the testList format)
            testentry = testList[testname]

            # we capture the path to the assembly file of this test
            test = testentry['test_path']
            test_name = os.path.splitext(os.path.basename(test))[0]

            # capture the directory where the artifacts of this test will be dumped/created.
            test_dir = testentry['work_dir']

            # name of the elf file after compilation of the test
            elf = test_name + '.elf'

            # name of the signature file as per requirement of RISCOF. RISCOF expects the signature to
            # be named as DUT-<dut-name>.signature. The below variable creates an absolute path of
            # signature file.
            sig_file = os.path.join(test_dir, self.name[:-1] + ".signature")

            # for each test there are specific compile macros that need to be enabled. The macros in
            # the testList node only contain the macros/values. For the gcc toolchain we need to
            # prefix with "-D". The following does precisely that.
            compile_macros= ' -D' + " -D".join(testentry['macros'])

            # collect the march string required for the compiler
            marchstr = testentry['isa'].lower()

            compile_cmd = self.compile_cmd.format(marchstr, self.xlen, test, elf, compile_macros)
            utils.shellCommand(compile_cmd).run(cwd=test_dir)

            objcopy_cmd = f'riscv32-elf-objcopy -O binary {elf} {test_name}.bin'
            utils.shellCommand(objcopy_cmd).run(cwd=test_dir)

            srec_cmd = f'srec_cat {test_name}.bin -binary -fill 0xFF -within {test_name}.bin -binary -range-pad 4 -byte-swap 4 -output {test_name}.hex -vmem 32'
            utils.shellCommand(srec_cmd).run(cwd=test_dir)

            if self.target_run:
                if not os.path.exists(self.dut_exe):
                    raise RuntimeError('DUT executable does not exist')

                # dut_run = [self.dut_exe, f'+initfile={test_name}.hex', f'+signature={sig_file}', '+waves']
                dut_run = [self.dut_exe, f'+initfile={test_name}.hex', f'+signature={sig_file}']
                utils.shellCommand(' '.join(dut_run)).run(cwd=test_dir)

        # if target runs are not required then we simply exit as this point after running all
        # the makefile targets.
        if not self.target_run:
            raise SystemExit
