module riscof_tb ();

  // Riscof tests can be big! RAM size of 8MB.
  parameter int RAM_DEPTH = 8192 * 1024 / 4;
  parameter int RAM_AW = $clog2(RAM_DEPTH);
  parameter int RAM_DW = 32;

  // --------------------------------------------------------------------------------
  // Logic Declarations
  // --------------------------------------------------------------------------------

  logic          clk;
  logic          rst;
  logic   [31:0] watchdog;
  logic   [31:0] i_addr;
  logic          i_rd;
  logic   [31:0] i_data;
  logic   [31:0] d_addr;
  logic   [31:0] d_data_s;
  logic   [31:0] d_data_l;
  logic   [ 3:0] d_data_s_stb;
  logic          d_store;
  logic          d_load;
  logic          d_load_done;
  logic          d_store_done;
  logic          dbg_enabled;
  logic          dbg_insn_ready;
  logic   [31:0] dbg_mbx_data;
  logic          fault;
  integer        fh;

  // --------------------------------------------------------------------------------
  // Clock/Reset
  // --------------------------------------------------------------------------------

  initial begin
    clk = '0;
    rst = '1;
    #10 rst = '0;
  end

  always #5 clk <= !clk;


  // --------------------------------------------------------------------------------
  // Testbench Control
  // --------------------------------------------------------------------------------

  initial begin
    if ($test$plusargs("waves")) begin
      $dumpfile("waves.fst");
      $dumpvars(0, riscof_tb);
    end
  end

  initial begin
    string signature;
    if ($value$plusargs("signature=%s", signature)) begin
      fh = $fopen(signature, "w");
    end
  end

  initial begin
    string filename;
    if ($value$plusargs("initfile=%s", filename)) begin
      $readmemh(filename, u_ram.mem);
    end
  end

  always @(posedge clk)
    if (rst) watchdog <= '0;
    else begin
      if (watchdog == 32'd10000000) begin
        $display("End of simulation due to watchdog timeout");
        $finish();
      end
      if (d_store && (d_addr == 32'hFFFFFFFC)) begin
        $display("End of test signalled by software");
        $finish();
      end
      if (d_store && (d_addr == 32'hFFFFFFF8)) begin
        $fdisplay(fh, "%x", d_data_s);
      end
      watchdog <= watchdog + 1'b1;
    end


  // --------------------------------------------------------------------------------
  // RAM Implementation
  // --------------------------------------------------------------------------------

  generic_ram #(
      .AW(RAM_AW),
      .DW(RAM_DW)
  ) u_ram (
      .clk,
      .cen0  (i_rd),
      .addr0 (i_addr[RAM_AW+2-1:2]),
      .rdata0(i_data),
      .cen1  (d_store || d_load),
      .wen1  (d_store),
      .addr1 (d_addr[RAM_AW+2-1:2]),
      .wdata1(d_data_s),
      .wstrb1(d_data_s_stb),
      .rdata1(d_data_l)
  );

  always @(posedge clk) begin
    if (rst) begin
      d_store_done <= '0;
      d_load_done  <= '0;
    end else begin
      d_store_done <= d_store;
      d_load_done  <= d_load;
    end
  end


  // --------------------------------------------------------------------------------
  // DUT
  // --------------------------------------------------------------------------------

  urv_cpu u_dut (
      .clk_i           (clk),
      .rst_i           (rst),
      .irq_i           ('0),
      .fault_o         (fault),
      .im_addr_o       (i_addr),
      .im_rd_o         (i_rd),
      .im_valid_i      ('1),
      .im_data_i       (i_data),
      .dm_addr_o       (d_addr),
      .dm_data_s_o     (d_data_s),
      .dm_data_l_i     (d_data_l),
      .dm_data_select_o(d_data_s_stb),
      .dm_store_o      (d_store),
      .dm_load_o       (d_load),
      .dm_load_done_i  (d_load_done),
      .dm_store_done_i (d_store_done),
      .dbg_force_i     ('0),
      .dbg_enabled_o   (dbg_enabled),
      .dbg_insn_i      ('0),
      .dbg_insn_set_i  ('0),
      .dbg_insn_ready_o(dbg_insn_ready),
      .dbg_mbx_data_i  ('0),
      .dbg_mbx_write_i ('0),
      .dbg_mbx_data_o  (dbg_mbx_data)
  );

  wire unused_ok = &{
    1'b0,
    fault,
    dbg_enabled,
    dbg_insn_ready,
    dbg_mbx_data,
    i_addr[1:0],
    i_addr[31:RAM_AW],
    1'b0
  };


endmodule
